from science_signal.signal import Signal
from science_signal import interpolate, interpolate_abciss
from numpy.typing import NDArray
from numpy import float64, zeros, pi, sqrt


def compute_light(
    scatter_factor: list[float],
    coupling: list[Signal],
    movement: Signal,
    wavelength: float,
    power_in: float,
    power_out: float,
) -> Signal:
    """
    Compute the projection from a given coupling and movement
    """
    frequencies = interpolate_abciss(
        movement.sin().psd().sqrt(), abs(coupling[0])
    )
    parts = zeros(
        (
            len(coupling),
            len(frequencies),
        )
    )

    for index in range(2):
        phase = (index + 1) * 4 * pi / wavelength

        factor_n = (movement * phase).sin().psd().sqrt()
        coupling_n = abs(coupling[0])
        factor_d = (movement * phase).cos().psd().sqrt()
        coupling_d = abs(coupling[1])

        factor_n, coupling_n, factor_d, coupling_d = interpolate(
            factor_n, coupling_n, factor_d, coupling_d
        )

        parts[index] = opt_compute_light(
            scatter_factor[index],
            factor_n,
            coupling_n,
            factor_d,
            coupling_d,
            power_in,
            power_out,
            phase,
        )

    return Signal(
        frequencies,
        sum(parts),
    )


def opt_compute_light(
    scatter_factor: float,
    factor_n: Signal,
    coupling_n: Signal,
    factor_d: Signal,
    coupling_d: Signal,
    power_in: float,
    power_out: float,
    phase: float,
) -> NDArray[float64]:
    """
    Optimized computing of light with pre-computed factor
    """
    return (
        sqrt(scatter_factor)
        / phase
        * power_in
        / power_out
        * (coupling_n * factor_n + coupling_d * factor_d).y
    )


def fit_compute_light(
    scatter_factor: float,
    factor_n: Signal,
    coupling_n: Signal,
    factor_d: Signal,
    coupling_d: Signal,
    power_in: float,
    power_out: float,
    data: Signal,
    reference: Signal,
    phase: float,
) -> float:
    """
    Scalar function used to find the right scattering factor
    """
    return sum(
        abs(
            Signal(
                factor_n.x,
                opt_compute_light(
                    scatter_factor=scatter_factor,
                    factor_n=factor_n,
                    coupling_n=coupling_n,
                    factor_d=factor_d,
                    coupling_d=coupling_d,
                    power_in=power_in,
                    power_out=power_out,
                    phase=phase,
                ),
            )
            + reference
            - data
        ).y
    )
